/*******************************************************************************
   Copyright 2008, 2009 Wearable Inc.; An Illinois, United States Corporation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*******************************************************************************/
/*******************************************************************************
   Name:        DOSonCHIP(TM) File Operation Definitions
   File Name:   DOSfileAPI.h
   Description: Defines the functions available for host software to call in
                order to interact with the DosOnChip system.
*******************************************************************************/

#ifndef DOSFILEAPI_H
#define DOSFILEAPI_H

#include "DOSprotocol.h"                           /* DOSonCHIP protocol */

/* CONSTANTS */
#define DOS_ERROR_RANGE_MAX                   0xDF    /* highest allowable error number */
#define DOS_HOST_ERROR_TIMEOUT                DOS_HANDSHAKE_OFF  /* host-side error if DOSonCHIP does fails to respond */
#define DOS_HOST_READ_NOT_COMPLETE            0xD0
#define DOS_HOST_WRITE_NOT_COMPLETE           0xD1
#define DOS_HOST_COMMUNICATION_ERROR          0xD2

/* DATA STRUCTURES */
/* Time & Date Structure */
struct DOStime
{
   unsigned char second; /* seconds after the minute: [0,59]  */
   unsigned char minute; /* minutes after the hour:   [0,59]  */
   unsigned char hour;   /* hours since midnight:     [0,23]  */
   unsigned char day;    /* day of the month:         [1,31]  */
   unsigned char month;  /* months since January:     [1,12]  */
   unsigned char year;   /* years since 1980:         [0,58]  */
};

/* GLOBAL VARIABLES */
/* The following holds command/response header + payload to/from DOSonCHIP IC */
extern unsigned char DOScommandResponse[DOS_PACKET_LENGTH_HEADER + DOS_PACKET_LENGTH_PAYLOAD_CMDRES];
/* The following variables are used to minimize commands sent to DOSonCHIP IC by mirroring last value sent */
extern unsigned char DOSpacketPayloadLengthName;   /* current DATA_NAME packet payload length */
extern unsigned int  DOSpacketPayloadLengthBlock;  /* current DATA_BLOCK packet payload length */
extern unsigned char DOScurrentHandle;             /* current handle */

/* FUNCTION PROTOTYPES */
void DOSdateTimeConvert (unsigned long int binaryDateTime, struct DOStime *dateTime);

unsigned char DOSsetName (unsigned char *name);    /* set the current file/directory name to DOSonCHIP */
unsigned char DOSgetName (unsigned char *name);    /* retrieve current file/directory name from DOSonCHIP */
unsigned char DOSsetHandle (unsigned char handle); /* set the current handle to DOSonCHIP */

unsigned char DOSmount (unsigned long int freeAllocateSize, unsigned long int *volumeID);

unsigned char DOSchangeDirectory (unsigned char *directory);
unsigned char DOSmakeDirectory (unsigned char *directory);
unsigned char DOSlistDirectory (unsigned char mode, unsigned char *attribute);
unsigned char DOSgetPropertyTime (unsigned char mode, struct DOStime *dateTime);
unsigned char DOSgetPropertySize (unsigned long int *size);

unsigned char DOSgetFreeSectorCount (unsigned long int *freeSectors);

unsigned char DOSopen (unsigned char mode, unsigned char *fileName, unsigned char handle, unsigned long int *numBytes);
unsigned char DOSwritePreallocate (unsigned char handle, unsigned long int numBytes, unsigned long int *fileSize);
unsigned char DOSclose (unsigned char handle);

unsigned char DOSseek (unsigned char handle, unsigned long int numBytes);
unsigned char DOSread (unsigned char handle, unsigned char *buffer, unsigned long int numBytes);
unsigned char DOSwrite (unsigned char handle, unsigned char *buffer, unsigned long int numBytes);

unsigned char DOSdelete (unsigned char *name);

unsigned char DOSsetTime (struct DOStime *dateTime);
unsigned char DOSgetTime (struct DOStime *dateTime);
unsigned char DOStimeOnOff (unsigned char switchState);

unsigned char DOSshutdown (void);
unsigned char DOSgetIC (unsigned char *versionIC);
unsigned char DOSgetVersion (unsigned int *versionBoot, unsigned int *versionFile);
unsigned char DOSupdateFirmware (void);

/* ERROR DIAGNOSTIC MACROS */
#define DOS_ERROR_BYTE DOScommandResponse[0]
#define DOS_ERROR_PARM_LONG                                 \
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] << 24 | \
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] << 16 | \
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] << 8  | \
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3]

/* BYTE-ACCESS DEFINES */

/* Define host byte order. Uncomment the appropriate definition for your architecture */
/* #define BIG_ENDIAN */
/* #define LITTLE_ENDIAN */

#if defined(BIG_ENDIAN)
/*
        Provides access to low byte of a two-byte word stored in big endian
        format (e.g. unsigned int).
*/
#define BYTELOW(v)   (*(((unsigned char *) (&v) + 1)))

/*
        Provides access to high byte of a two-byte word stored in big endian
        format (e.g. unsigned int).
*/
#define BYTEHIGH(v)  (*((unsigned char *) (&v)))

/*
        Provides access to high byte of high word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDHIGH_BYTEHIGH(v)   (*((unsigned char *) (&v)))

/*
        Provides access to low byte of high word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDHIGH_BYTELOW(v)   (*(((unsigned char *) (&v) + 1)))

/*
        Provides access to high byte of low word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDLOW_BYTEHIGH(v)   (*(((unsigned char *) (&v) + 2)))

/*
        Provides access to high byte of high word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDLOW_BYTELOW(v)   (*(((unsigned char *) (&v) + 3)))

/*
        Provides access to low word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDLOW(v)   (*(((unsigned int *) (&v) + 1)))

/*
        Provides access to low word of a four-byte double word 
        stored in big endian format (e.g. unsigned long int).
*/
#define WORDHIGH(v)   (*((unsigned int *) (&v)))

#elif defined(LITTLE_ENDIAN)

/*
        Provides access to low byte of a two-byte word stored in little endian
        format (e.g. unsigned int).
*/
#define BYTELOW(v)   (*(((unsigned char *) (&v))))

/*
        Provides access to high byte of a two-byte word stored in little endian
        format (e.g. unsigned int).
*/
#define BYTEHIGH(v)  (*((unsigned char *) (&v) + 1))

/*
        Provides access to high byte of high word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDHIGH_BYTEHIGH(v)   (*((unsigned char *) (&v) + 3))

/*
        Provides access to low byte of high word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDHIGH_BYTELOW(v)   (*(((unsigned char *) (&v) + 2)))

/*
        Provides access to high byte of low word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDLOW_BYTEHIGH(v)   (*(((unsigned char *) (&v) + 1)))

/*
        Provides access to high byte of high word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDLOW_BYTELOW(v)   (*(((unsigned char *) (&v))))

/*
        Provides access to low word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDLOW(v)   (*(((unsigned int *) (&v))))

/*
        Provides access to low word of a four-byte double word 
        stored in little endian format (e.g. unsigned long int).
*/
#define WORDHIGH(v)   (*((unsigned int *) (&v) + 1))

#else

#error One of BIG_ENDIAN or LITTLE_ENDIAN must be defined.

#endif

#endif
