/*******************************************************************************
   Copyright 2008, 2009 Wearable Inc.; An Illinois, United States Corporation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*******************************************************************************/
/*******************************************************************************
   Name:        DOSonCHIP(TM) Packet Communications Functions
   File Name:   DOSpacket.c
   Description: Functions to communicate packets between host and DOSonCHIP IC.
*******************************************************************************/

/* FILE INCLUDES */
#include "DOSprotocol.h"
#include "DOSfileapi.h"
#include "DOSpacket.h"
#include "DOSport.h"

/*******************************************************************************/
unsigned char DOSpacketPut (unsigned char header, unsigned char *payload) /* Transmit packet to DOSonCHIP */
/*
   Input:   header  : packet header byte
            payload : pointer to memory buffer containing the payload
   Returns: 0|error
   Note:    the error/non-error result is saved for error diagnostics; the error detail parameter is cleared
*/
{
   unsigned char err;
   unsigned int  len;
   
   DOS_PORT_ENABLE;                                /* enable DOSonCHIP communications port */

   err = DOSportGetNotBusy ();

   if (err == DOS_HANDSHAKE_GO)                    /* need DOS_HANDSHAKE_GO from DOSonCHIP to send packet */
   {
      len = 0;                                     /* assume error: no packet payload */

      DOSportPutByte (header);                     /* send packet header */

      /* Set packet payload length */
      switch (header)
      {
         case DOS_HANDSHAKE_PAK_NEXT:
            /* len = DOS_PACKET_LENGTH_PAYLOAD_HANDSHAKE; */
            /* len = 0; */
            break;
         case DOS_HANDSHAKE_PAK_ABORT:
            /* len = DOS_PACKET_LENGTH_PAYLOAD_HANDSHAKE; */
            /* len = 0; */
            break;
         case DOS_DATA_NAME:
            len = DOSpacketPayloadLengthName;      /* current packet payload length */
            break;
         case DOS_DATA_BLOCK:
            len = DOSpacketPayloadLengthBlock;     /* current packet payload length */
            break;
         default:
            /* len = DOS_PACKET_LENGTH_PAYLOAD_CMDRES; */
            /* Send using global variables (ignore "payload" pointer) */
            DOSportPutByte (DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0]);
            DOSportPutByte (DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1]);
            DOSportPutByte (DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2]);
            DOSportPutByte (DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3]);
            /* len = 0; */
            break;
      }

      /* Transfer payload */
      while (len > 0)
      {
         DOSportPutByte (*(payload++));
         len--;
      }

      /* No footer */
     
     err = 0;                                     /* no error */
   }

   DOSportFlushQ ();                               /* Make sure transmit/receive queue is clear */
   
   DOS_PORT_DISABLE;                               /* disable DOSonCHIP communications port */

   /* Save error for diagnostics */
   DOScommandResponse[0] = err;                    /* packet header = 0|error */
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] = 0; /* Clear error detail */
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3] = 0;
   
   return (DOScommandResponse[0]);
}

/*******************************************************************************/
void DOSpacketPutBlock (unsigned char *payload)    /* Transmit DOS_DATA_BLOCK packet to DOSonCHIP */
/*
   Input:   payload : pointer to memory buffer containing the payload
   Returns: void
*/
{
   unsigned int  len;
   
   DOS_PORT_ENABLE;                                /* enable DOSonCHIP communications port */

   DOSportPutByte (DOS_DATA_BLOCK);               /* send packet header */
   
   len = DOSpacketPayloadLengthBlock;              /* current packet payload length */

   /* Transfer payload */
   while (len > 0)
   {
      DOSportPutByte (*(payload++));
      len--;
   }

   /* No footer */

   DOSportFlushQ ();                               /* Make sure transmit/receive queue is clear */
   
   DOS_PORT_DISABLE;                               /* disable DOSonCHIP communications port */
}

/*******************************************************************************/   
unsigned char DOSpacketGet (unsigned char *payload) /* Receive packet from DOSonCHIP */
/*
   Input:   payload : pointer to memory buffer to place the received payload
   Returns: packet header byte | error
   Note:    if command or response, payload is placed into global array
*/
{
   unsigned int  len = 0;
   unsigned char response = 0;                     /* assume no error */

   /* Clear error detail */
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3] = 0;
     
   DOS_PORT_ENABLE;                                /* enable DOSonCHIP communications port */
   
   DOScommandResponse[0] = DOSportGetNotBusy ();   /* get packet header */
   response = DOScommandResponse[0];
   switch (DOScommandResponse[0])
   {
      case DOS_HANDSHAKE_WAIT:
        /* DOScommandResponse[0] = DOS_HOST_ERROR_TIMEOUT; */
        response = DOS_HOST_ERROR_TIMEOUT;
        /* len = 0; */
        break;
      case DOS_HANDSHAKE_OFF:
        /* DOScommandResponse[0] = DOS_HOST_COMMUNICATION_ERROR; */
        response = DOS_HOST_COMMUNICATION_ERROR;
        /* len = 0; */
        break;
      case DOS_DATA_NAME:
         len = DOSpacketPayloadLengthName;         /* current packet payload length */
         break; 
      case DOS_DATA_BLOCK:
         len = DOSpacketPayloadLengthBlock;        /* current packet payload length */
         break;
      default:
         /* if command/response packet, save for error diagnostics */
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3] = DOSportGetByte ();
         /* len = 0; */
         break;
   }

   /* Transfer payload, if any */
   while (len > 0)
   {
      *(payload++) = DOSportGetByte ();
      len--;
   }
   
   /* No footer */

   DOS_PORT_DISABLE;                               /* disable DOSonCHIP communications port */

   return (response);                              /* return packet header | error */
}

/*******************************************************************************/   
unsigned char DOSpacketGetResponse (void) /* Receive a response packet from DOSonCHIP */
/*
   Returns: packet header byte | error
   Note:    if response, payload is placed into global array
*/
{
   unsigned int  len = 0;
   unsigned char response = 0;                     /* assume no error */

   /* Clear error detail */
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] = 0;
   DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3] = 0;
     
   DOS_PORT_ENABLE;                                /* enable DOSonCHIP communications port */
   
   DOScommandResponse[0] = DOSportGetNotBusy ();   /* get packet header */
   response = DOScommandResponse[0];
   switch (DOScommandResponse[0])
   {
      case DOS_HANDSHAKE_GO:
         /* len = 0; */
         break;
      case DOS_HANDSHAKE_WAIT:
         /* DOScommandResponse[0] = DOS_HOST_ERROR_TIMEOUT; */
         response = DOS_HOST_ERROR_TIMEOUT;
         /* len = 0; */
         break;
      case DOS_HANDSHAKE_OFF:
         /* DOScommandResponse[0] = DOS_HOST_COMMUNICATION_ERROR; */
         response = DOS_HOST_COMMUNICATION_ERROR;
         /* len = 0; */
         break;
      case DOS_DATA_NAME:
         response = DOS_RES_INVALID_PACKET_TYPE;
         len = DOSpacketPayloadLengthName;         /* current packet payload length */
         break; 
      case DOS_DATA_BLOCK:
         response = DOS_RES_INVALID_PACKET_TYPE;
         len = DOSpacketPayloadLengthBlock;        /* current packet payload length */
         break;
      default:
         /* get response packet payload */
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 0] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 1] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 2] = DOSportGetByte ();
         DOScommandResponse[DOS_PACKET_LENGTH_HEADER + 3] = DOSportGetByte ();
         /* len = 0; */
         break;
   }

   /* Absorb improper payload from DOSonCHIP IC if improper data packet */
   while (len > 0)
   {
      DOSportGetByte ();
      len--;
   }
   
   /* No footer */

   DOS_PORT_DISABLE;                               /* disable DOSonCHIP communications port */

   return (response);                              /* return packet header | error */
}
