/*******************************************************************************
   Copyright 2008, 2009 Wearable Inc.; An Illinois, United States Corporation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*******************************************************************************/
/*******************************************************************************
   Name:        DOSonCHIP(TM) Example "Main" Code
   File Name:   main.c
   Description: An example program for DOSonCHIP API/SDK.
*******************************************************************************/

/**************************************************************************************************/
/* FILE INCLUDES */
#include "DOSONCHIP/DOSprotocol.h"         /* DOSonCHIP protocol */
#include "DOSONCHIP/DOSfileAPI.h"          /* DOSonCHIP API */
#include "DOSONCHIP/DOSport_SPI_BitBang.h" /* DOSonCHIP (function-prototype for DOSportInit) */
#include "IC_Init.h"                       /* CPU specific init routine(s); user supplied */
#include "string.h"                        /* for string operations */
#include "stdio.h"                         /* for diagnostic output (printf) */

/**************************************************************************************************/
/* MACROS */
/* Error Handler */
#define ERROR_CHECK(response)                                              \
   do                                                                      \
   {                                                                       \
      if (response) /* if error, print error data & loop forever */        \
      {                                                                    \
         printf ("ERROR %BX: %LX\n", DOS_ERROR_BYTE, DOS_ERROR_PARM_LONG); \
         while (1); /* infinite loop */                                    \
      }                                                                    \
   } while (0)

/**************************************************************************************************/
/**
   main () routine.
*/
/**************************************************************************************************/
void main (void) 
{
   unsigned char     versionIC;
   unsigned int      versionBoot, versionFile;
   struct   DOStime  dateTime;
   unsigned long int idata volumeID;
   unsigned char     handle;
   unsigned long int idata fileSize, length;
   unsigned long int idata numBytes;
   unsigned char     attribute;
   unsigned char     idata name[13];        /* 8.3 plus null */
   unsigned char     idata dataBuffer[32];  /* create memory buffer */
   unsigned char     status;

   /* Initialize host IC & its communication port to DOSonCHIP */
   InitIC ();      /* initialize host (CPU specific; user supplied) */
   DOSportInit (); /* initialize host communications port (CPU specific) */

   printf ("\n");

   /***********************************************************************************************/
   /* Prepare DOSonCHIP for a firwmare update */
   /* status = DOSupdateFirmware ();
   ERROR_CHECK (status);
   */

   /***********************************************************************************************/
   /* Get DOSonCHIP information */
   status = DOSgetIC (&versionIC);
   ERROR_CHECK (status);
   printf ("DOSonCHIP IC Version:          %02BX\n", versionIC);
   
   status = DOSgetVersion (&versionBoot, &versionFile);
   ERROR_CHECK (status);
   printf ("DOSonCHIP IC Boot Version:     %04X\n", versionBoot);
   printf ("DOSonCHIP IC Firmware Version: %04X\n\n", versionFile);

   /***********************************************************************************************/
   /* Set date & time */
   dateTime.year   = 2009 - 1980;   /* must subtract 1980 from year */
   dateTime.month  = 10;
   dateTime.day    = 31;
   dateTime.hour   = 23;
   dateTime.minute = 59;
   dateTime.second = 00;

   status = DOSsetTime (&dateTime); /* Set DOSonCHIP date & time for file time stamping */
   ERROR_CHECK (status);

   printf ("DOSonCHIP Date & Time: %04u-%02bu-%02bu %02bu:%02bu:%02bu\n\n",
           dateTime.year+1980, dateTime.month, dateTime.day,
           dateTime.hour, dateTime.minute, dateTime.second);
   
   status = DOStimeOnOff (DOS_ON);  /* Enable DOSonCHIP real-time clock */
   ERROR_CHECK (status);

   /***********************************************************************************************/   
   /* Mount card file volume */
   /* Cache maximum allowable unused sectors for write acceleration */
   /* & return unique card volume ID */
   status = DOSmount (0xFFFFFFFF, &volumeID);
   ERROR_CHECK (status);

   /***********************************************************************************************/   
   /* Get file volume free sector space */
   /* Get the total number of free sectors available for file system use on the mounted card */
   status = DOSgetFreeSectorCount (&length);
   ERROR_CHECK (status);
   printf ("Free Sectors: %lu\n\n", length);

   /***********************************************************************************************/   
   /* Make new directory */
   /* Initially in root directory (from DOSmount command) */
   status = DOSmakeDirectory ("log"); /* DOSonCHIP will convert this to upper case */

   if (status == DOS_RES_NAME_DUPLICATE)
      printf ("Directory 'LOG' is already present\n");
   ERROR_CHECK (status);

   /***********************************************************************************************/   
   /* Change current directory */
   status = DOSchangeDirectory ("LOG"); /* previously in root directory */
   ERROR_CHECK (status);

   /***********************************************************************************************/
   /* Open file for writing */
   strcpy (name, "test.txt"); /* file name */
   handle = 0; /* assign to handle 0 */
   status = DOSopen (DOS_CMD_OPEN_WRITE, &name, handle, &fileSize);
   ERROR_CHECK (status);

   /* Increase file size in advance to minimize delay between DOSwrite commands */
   numBytes = 8 * 1024 * 1024; /* 8 Mbytes */
   status = DOSwritePreallocate (handle, numBytes, &fileSize); /* increase file size by numBytes */
   ERROR_CHECK (status);   
   
   /* Write to file */
   strcpy (dataBuffer, "say Hello");
   
   status = DOSwrite (handle, &dataBuffer, strlen (dataBuffer));
   ERROR_CHECK (status);
   
   status = DOSwrite (handle, " to the World!", 14);
   ERROR_CHECK (status);

   /* Close file */
   status = DOSclose (handle);
   ERROR_CHECK (status);

   /***********************************************************************************************/
   /* Open file for reading */
   status = DOSopen (DOS_CMD_OPEN_READ, name, 1, &fileSize); /* file size is returned */
   length = fileSize; /* save length of file for later */
   ERROR_CHECK (status);

   /* Open file for reading */
   handle = 2; /* assign to handle 2 */
   status = DOSopen (DOS_CMD_OPEN_READ, name, handle, &fileSize); /* file size is returned */
   ERROR_CHECK (status);
   
   /* Read from file */
   status = DOSread (1, &dataBuffer, length); /* handle 1 */
   ERROR_CHECK (status);

   dataBuffer[length] = 0; /* add null terminator for string handling */
   printf ("%s\n", dataBuffer);   

   /* Seek to later part in the file */
   /* Move the file byte pointer forward relative to the current position */
   status = DOSseek (handle, 4); /* skip 4 bytes */
   ERROR_CHECK (status);

   /* Read from file */
   status = DOSread (handle, &dataBuffer, 5);
   ERROR_CHECK (status);
   
   /* Seek to later part in the file */
   /* Move the file byte pointer forward relative to the current position */
   status = DOSseek (handle, 7); /* skip 7 bytes */
   ERROR_CHECK (status);
   
   /* Read from file */
   status = DOSread (handle, &dataBuffer[5], 7);
   ERROR_CHECK (status);

   dataBuffer[12] = 0; /* add null terminator for string handling */
   printf ("%s\n", dataBuffer);
   
   /* Close files */
   status = DOSclose (handle);
   ERROR_CHECK (status);

   status = DOSclose (1); /* handle 1 */
   ERROR_CHECK (status);
   
   /***********************************************************************************************/
   /* List directory contents */
   printf ("\n%-12s %6s %-19s %-19s\n", "NAME", "SIZE", "CREATED", "MODIFIED");
   status = DOSlistDirectory (DOS_FIRST, &attribute);
   if (status != DOS_RES_DIR_END)
   {
      while (!status)
      {
         status = DOSgetName (&name);
         ERROR_CHECK (status);
         printf ("%-12s", name);
         
         if (attribute & DOS_ATTR_DIRECTORY)
         {
            printf (" %6s", "\\");
         }
         else
         {
            status = DOSgetPropertySize (&fileSize);
            ERROR_CHECK (status);
            printf (" %6lu", fileSize);
         }
         
         status = DOSgetPropertyTime (DOS_PROPERTY_TIME_CREATED, &dateTime);
         ERROR_CHECK (status);
         printf (" %04u-%02bu-%02bu %02bu:%02bu:%02bu",
                 dateTime.year+1980, dateTime.month, dateTime.day,
                 dateTime.hour, dateTime.minute, dateTime.second);
         
         status = DOSgetPropertyTime (DOS_PROPERTY_TIME_MODIFIED, &dateTime);
         ERROR_CHECK (status);
         printf (" %04u-%02bu-%02bu %02bu:%02bu:%02bu",
                 dateTime.year+1980, dateTime.month, dateTime.day,
                 dateTime.hour, dateTime.minute, dateTime.second);
         
         printf ("\n");
         status = DOSlistDirectory (DOS_NEXT, &attribute);
      }
      if (status != DOS_RES_DIR_END)
      {
         ERROR_CHECK (status);
      }
   }
   printf ("\n");

   /***********************************************************************************************/
   /* Delete file */
   status = DOSdelete (&name);
   ERROR_CHECK (status);
   
   /* Delete all remaining files in current directory */
   while (!status)
   {
      status = DOSlistDirectory (DOS_FIRST, &attribute);
      if (status == DOS_RES_DIR_END) break;
      ERROR_CHECK (status);
      if (!(attribute & DOS_ATTR_DIRECTORY))
      {
         status = DOSgetName (&name);
         ERROR_CHECK (status);

         status = DOSdelete (&name);
         ERROR_CHECK (status);
      }
   }

   /***********************************************************************************************/   
   /* Change current directory */
   status = DOSchangeDirectory ("\\"); /* jump to root directory */
   ERROR_CHECK (status);

   /***********************************************************************************************/
   /* Get date & time */
   status = DOSgetTime (&dateTime); /* Get current DOSonCHIP date & time */
   ERROR_CHECK (status);

   printf ("DOSonCHIP Date & Time: %04u-%02bu-%02bu %02bu:%02bu:%02bu\n\n",
           dateTime.year+1980, dateTime.month, dateTime.day,
           dateTime.hour, dateTime.minute, dateTime.second);

   /***********************************************************************************************/   
   /* Shut down DOSonCHIP IC */
   printf ("DOSonCHIP Shutdown\n");
   status = DOSshutdown ();

   ERROR_CHECK (status);

   /***********************************************************************************************/ 
   /* End of main.c: loop forever */
   while (!status);
}